{-# OPTIONS
  -XFlexibleInstances
  -XMultiParamTypeClasses
  -XFunctionalDependencies
  -XOverlappingInstances
  -XUndecidableInstances
#-}

module Monatron.AutoLift (
 StateM(..), get,put,
 WriterM (..), tell,collect,listen,pass,
 ReaderM(..), ask,local,
 ExcM(..), throw,handle,
 ContM(..), callCC,
 PlusM(..), mZero,mPlus,
 module Monatron.Transformer
) where

import Monatron.Transformer
import Control.Exception (SomeException)


------------------------------------------------------------------
-- MakeWith for State
class Monad m => StateM z m | m -> z where
    stateM :: MakeWith z m

instance Monad m => StateM z (StateT z m) where
    stateM = makeWithStateT

instance (StateM z m, MonadT t) => StateM z (t m) where
    stateM = liftMakeWith stateM

get :: StateM z m => m z
get = getX stateM

put :: StateM z m => z -> m ()
put = putX stateM

------------------------------------------------------------------
-- With for Output
class (Monoid z, Monad m) => WriterM z m | m -> z where
    writerM :: MakeWith z m

instance (Monoid z, Monad m) => WriterM z (WriterT z m) where
    writerM = makeWithWriterT

instance (Monoid z, WriterM z m, MonadT t) => 
          WriterM z (t m) where
    writerM = liftMakeWith writerM

tell :: (Monoid z, WriterM z m) => z -> m ()
tell z = putX writerM z

collect :: (WriterM z m) => m a -> m (a,z)
collect = collectX writerM

listen :: (WriterM z m) => m a -> m (a,z)
listen m = collect m >>= \(a,z) -> tell z >> return (a,z)

pass :: (WriterM z m) => m (a, z -> z) -> m a
pass m = collect m >>= \((a, f),z) -> tell (f z) >> return a

------------------------------------------------------------------
-- With for Environments
class Monad m => ReaderM z m | m -> z where
    readerM :: MakeWith z m

instance Monad m => ReaderM z (ReaderT z m) where
    readerM = makeWithReaderT

instance (ReaderM z m, MonadT t) => ReaderM z (t m) where
    readerM = liftMakeWith readerM

ask :: ReaderM z m => m z
ask = getX readerM

local :: ReaderM z m => (z -> z) -> m a -> m a
local = localX readerM

------------------------------------------------------------------
-- Throw and Catch for Exceptions

class Monad m => ExcM z m | m -> z where
    excM :: Exc z m

instance Monad m => ExcM z (ExcT z m) where
    excM = excExcT

instance ExcM SomeException IOExc where
    excM = excIOExc

instance (ExcM z m, MonadT t) => ExcM z (t m) where
    excM = liftExc excM

throw :: ExcM z m => z -> m a
throw = throwX excM

handle :: ExcM z m => m a -> (z -> m a) -> m a
handle = handleX excM

------------------------------------------------------------------
-- callCC operation

class Monad m => ContM m where
    callCCM :: CallCC m

instance Monad m => ContM (ContT r m) where
    callCCM = callCCContT

instance (ContM m, MonadT t) => ContM (t m) where
    callCCM = liftCallCC callCCM

callCC :: ContM m => ((a -> m b) -> m a) -> m a
callCC = callCCX callCCM

------------------------------------------------------------------
-- MPlus operations

class Monad m => PlusM m where
    plusM :: MPlus m

instance Monad m => PlusM (ListT m) where
    plusM = mPlusListT

instance (PlusM m, MonadT t) => PlusM (t m) where
    plusM = liftMPlus plusM

mZero :: (PlusM m) => m a
mZero = mZeroX plusM

mPlus :: PlusM m => m a -> m a -> m a
mPlus = mPlusX plusM

-- instance for the standard list monad
instance PlusM [] where
    plusM = mPlusList
