{-# OPTIONS -XImpredicativeTypes -XRank2Types #-}

module Monatron.Operations (
  MakeWith,  MakeWithSig(..), liftMakeWith,
  withX, getX, collectX, makeX, putX, localX,
  Exc, ExcSig(..), liftExc, throwX, handleX,
  CallCC, liftCallCC, callCCX, 
  MPlus, MPlusSig(..), liftMPlus, mZeroX, mPlusX, 
  module Monatron.MonadT
) where

import Monatron.MonadT

type Norm m = forall a. m a -> m a

--------------------------------------------------------
data MakeWithSig z a = With (z -> a)
                     | Make a z

type MakeWith z m = (forall a. MakeWithSig z a -> m a, Norm m) 



withX :: Monad m => MakeWith z m -> (z -> m a) -> m a
withX (k,n) = n . join . k . With      

makeX :: Monad m => MakeWith z m -> m a -> z -> m a
makeX (k,n)  m z = n $ join $ k $ Make m z

liftMakeWith :: (Monad m, MonadT t) => MakeWith z m -> MakeWith z (t m)
liftMakeWith (k,n) = (lift . k, tmap n)

-- derived:
getX :: Monad m => MakeWith z m -> m z
getX op = withX op return

putX :: Monad m => MakeWith z m -> z -> m ()
putX op = makeX op (return ())

collectX :: Monad m => MakeWith z m -> m a -> m (a,z)
collectX op m = m >>= \a -> withX op (\z -> return (a,z)) 

localX :: Monad m => MakeWith z m -> (z -> z) -> m a -> m a
localX op f m = do z <- getX op
                   makeX op m (f z)

--------------------------------------------------------
data ExcSig z a = Throw z
                | Handle a (z -> a)

type Exc z m = (forall a. ExcSig z a -> m a, Norm m)

throwX :: Monad m => Exc z m -> z -> m a
throwX (k,n)  = n . k . Throw

handleX :: Monad m => Exc z m -> m a -> (z -> m a) -> m a
handleX (k,n) m  =  n . join . k . Handle m 

liftExc :: (Monad m, MonadT t) => Exc z m -> Exc z (t m)
liftExc (k,n)  = (lift . k, tmap n)

--------------------------------------------------------
type CallCC m = (forall a b. ((a -> m b) -> a) -> m a, Norm m)

callCCX :: (Monad m) => CallCC m -> ((a -> m b) -> m a) -> m a
callCCX (k,n) f = n $ join $ k (\c -> f (c . return))

liftCallCC :: (Monad m, MonadT t) => CallCC m -> CallCC (t m)
liftCallCC (k,n)  = (\f -> lift $ k (\c -> f (lift. c)), tmap n)

--------------------------------------------------------
data MPlusSig a = MZero | MPlus a a

type MPlus m = (forall a. MPlusSig a -> m a, Norm m)

mZeroX             :: Monad m => MPlus m -> m a
mZeroX (k,n)  = n (k MZero)

mPlusX      :: Monad m => MPlus m -> m a -> m a -> m a
mPlusX (k,n)  m =  n . join . k . MPlus m

liftMPlus :: (Monad m, MonadT t) => MPlus m -> MPlus (t m)
liftMPlus (k,n)  = (lift . k, tmap n)
